# Generated by Django 2.1.1 on 2018-10-04 12:19

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("bankreader", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Order",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=30)),
                ("variable_symbol", models.CharField(max_length=10, unique=True)),
            ],
            options={
                "verbose_name": "order",
                "verbose_name_plural": "orders",
            },
        ),
        migrations.CreateModel(
            name="OrderPayment",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "amount",
                    models.DecimalField(decimal_places=2, max_digits=20, verbose_name="amount"),
                ),
                (
                    "order",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="demoapp.Order"),
                ),
                (
                    "transaction",
                    models.OneToOneField(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="identified_order_payment",
                        to="bankreader.Transaction",
                    ),
                ),
            ],
            options={
                "verbose_name": "order payment",
                "verbose_name_plural": "order payments",
            },
        ),
    ]



from typing import Iterable, List

from django.db import IntegrityError, models, transaction as db_transaction
from django.utils.translation import gettext, gettext_lazy as _
from localflavor.generic.models import BICField, IBANField

from .readers import get_reader_choices, readers
from .readers.base import BaseReader


class Account(models.Model):
    name = models.CharField(_("account name"), max_length=150, unique=True)
    iban = IBANField(_("IBAN"), blank=True, null=True)
    bic = BICField(_("BIC (SWIFT)"), blank=True, null=True)
    reader = models.CharField(
        _("account statement format"),
        blank=True,
        choices=get_reader_choices(),
        max_length=150,
        null=True,
    )

    class Meta:
        verbose_name = _("account")
        verbose_name_plural = _("accounts")

    def __str__(self) -> str:
        return self.name

    def get_reader(self) -> BaseReader | None:
        return readers.get(self.reader) if self.reader is not None else None


class AccountStatement(models.Model):
    account = models.ForeignKey(
        Account,
        limit_choices_to={"reader__isnull": False},
        on_delete=models.CASCADE,
        related_name="account_statements",
        verbose_name=_("account"),
    )
    statement = models.CharField(_("statement"), max_length=256)
    from_date = models.DateField(_("from date"), editable=False)
    to_date = models.DateField(_("to date"), editable=False)

    class Meta:
        ordering = ("from_date",)
        verbose_name = _("account statement")
        verbose_name_plural = _("account statements")

    def __str__(self) -> str:
        return self.statement

    @db_transaction.atomic
    def save_with_transactions(self, transactions: Iterable["Transaction"]) -> List[str]:
        self.from_date = min(td.accounted_date for td in transactions)
        self.to_date = max(td.accounted_date for td in transactions)
        self.save()
        messages = []
        for transaction in transactions:
            try:
                with db_transaction.atomic():
                    transaction.account = self.account
                    transaction.account_statement = self
                    transaction.save()
            except IntegrityError:
                messages.append(
                    gettext('Transaction "{transaction_id}" already exists for account "{account_name}".').format(
                        transaction_id=transaction.transaction_id,
                        account_name=self.account.name,
                    )
                )
        return messages


class Transaction(models.Model):
    transaction_id = models.CharField(_("transaction id"), max_length=256)
    account_statement = models.ForeignKey(
        AccountStatement,
        on_delete=models.CASCADE,
        related_name="transactions",
        verbose_name=_("account statement"),
    )
    account = models.ForeignKey(Account, on_delete=models.CASCADE, verbose_name=_("account"))
    entry_date = models.DateField(_("entry date"))
    accounted_date = models.DateField(_("accounted date"))
    remote_account_number = models.CharField(_("remote account number"), max_length=64)
    remote_account_name = models.CharField(_("remote account name"), max_length=128)
    amount = models.DecimalField(_("amount"), decimal_places=2, max_digits=20)
    variable_symbol = models.BigIntegerField(_("variable symbol"), default=0)
    constant_symbol = models.BigIntegerField(_("constant symbol"), default=0)
    specific_symbol = models.BigIntegerField(_("specific symbol"), default=0)
    sender_description = models.CharField(_("description for sender"), max_length=256)
    recipient_description = models.CharField(_("description for recipient"), max_length=256)

    class Meta:
        ordering = ("accounted_date",)
        unique_together = ("account", "transaction_id")
        verbose_name = _("transaction")
        verbose_name_plural = _("transactions")

    def __str__(self) -> str:
        return f"{self.accounted_date} {self.amount} {self.remote_account_name} {self.sender_description}"



import logging
from typing import TYPE_CHECKING, Any, Callable, Dict, Generator, List, Tuple, Type

from django import forms
from django.contrib import admin, messages
from django.contrib.admin.views.main import ChangeList
from django.core.exceptions import ValidationError
from django.core.files.uploadedfile import UploadedFile
from django.db import models
from django.db.models.fields.reverse_related import OneToOneRel
from django.http import HttpRequest
from django.templatetags.static import static
from django.urls import reverse_lazy as reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from .models import Account, AccountStatement, Transaction

logger = logging.getLogger(__name__)


if TYPE_CHECKING:

    class AccountWithStatementsCount(Account):
        account_statements_count: int

    class AccountStatementWithTransactionsCount(AccountStatement):
        transactions_count: int


def get_transaction_relations() -> Dict[str, OneToOneRel]:
    return {rel.name: rel for rel in Transaction._meta.related_objects if isinstance(rel, OneToOneRel)}  # type: ignore


class AmountFieldListFilter(admin.FieldListFilter):
    def __init__(
        self,
        field: models.Field,
        request: HttpRequest,
        params: Dict[str, str],
        model: Type[models.Model],
        model_admin: admin.ModelAdmin,
        field_path: str,
    ) -> None:
        self.lookup_kwarg_credit = "%s__gt" % field_path
        self.lookup_kwarg_debit = "%s__lt" % field_path
        self.lookup_val_credit = params.get(self.lookup_kwarg_credit)
        self.lookup_val_debit = params.get(self.lookup_kwarg_debit)
        super().__init__(field, request, params, model, model_admin, field_path)

    def expected_parameters(self) -> list[str | None]:
        return [self.lookup_kwarg_credit, self.lookup_kwarg_debit]

    def choices(self, changelist: ChangeList) -> Generator[dict[str, Any], None, None]:
        yield {
            "selected": self.lookup_val_credit is None and self.lookup_val_debit is None,
            "query_string": changelist.get_query_string(remove=[self.lookup_kwarg_credit, self.lookup_kwarg_debit]),
            "display": _("All"),
        }
        yield {
            "selected": self.lookup_val_credit == "0",
            "query_string": changelist.get_query_string({self.lookup_kwarg_credit: "0"}, [self.lookup_kwarg_debit]),
            "display": _("Credit transactions"),
        }
        yield {
            "selected": self.lookup_val_debit == "0",
            "query_string": changelist.get_query_string({self.lookup_kwarg_debit: "0"}, [self.lookup_kwarg_credit]),
            "display": _("Debit transactions"),
        }


class IdentifiedFieldListFilter(admin.RelatedFieldListFilter):
    def choices(self, changelist: ChangeList) -> Generator[dict[str, Any], None, None]:
        yield {
            "selected": self.lookup_val_isnull is None,
            "query_string": changelist.get_query_string(remove=[self.lookup_kwarg_isnull]),
            "display": _("All"),
        }
        yield {
            "selected": self.lookup_val_isnull == "True",
            "query_string": changelist.get_query_string({self.lookup_kwarg_isnull: "True"}),
            "display": _("Not identified"),
        }
        yield {
            "selected": self.lookup_val_isnull == "False",
            "query_string": changelist.get_query_string({self.lookup_kwarg_isnull: "False"}),
            "display": _("Identified"),
        }


@admin.register(Account)
class AccountAdmin(admin.ModelAdmin):
    list_display = ("name", "iban", "bic", "account_statements_link")

    def get_queryset(self, request: HttpRequest) -> models.QuerySet["AccountWithStatementsCount"]:
        return super().get_queryset(request).annotate(account_statements_count=models.Count("account_statements"))

    account_statement_changelist = reverse("admin:bankreader_accountstatement_changelist")

    @admin.display(description=_("account statements"), ordering="account_statements_count")
    def account_statements_link(self, obj: "AccountWithStatementsCount") -> str:
        return mark_safe(
            '<a href="{url}?account__id__exact={account_id}">{count}</a>'.format(
                url=self.account_statement_changelist,
                account_id=obj.id,
                count=obj.account_statements_count,
            )
        )


class ReadOnlyMixin:
    def has_change_permission(self, request: HttpRequest, obj: models.Model | None = None) -> bool:
        return False


class AccountStatementForm(forms.ModelForm):
    statement = forms.FileField(label=_("account statement"))
    transactions: tuple[Transaction, ...] | None = None

    def clean(self) -> dict[str, Any]:
        account: Account | None = self.cleaned_data.get("account")
        statement: UploadedFile | None = self.cleaned_data.get("statement")
        if account is None or statement is None or statement.file is None:
            return self.cleaned_data
        reader = account.get_reader()
        assert reader is not None
        try:
            self.transactions = tuple(reader.read_file(statement.file))
        except Exception:
            msg = _("Failed to read transaction data in format {}.").format(reader.label)
            logger.exception(msg)
            raise ValidationError(msg)
        if not self.transactions:
            raise ValidationError(_("The account statement doesn't contain any transaction data."))
        return self.cleaned_data


@admin.register(AccountStatement)
class AccountStatementAdmin(ReadOnlyMixin, admin.ModelAdmin):
    form = AccountStatementForm
    list_display = (
        "id",
        "statement",
        "account_name",
        "from_date",
        "to_date",
        "transactions_link",
    )
    list_filter = ("account",)
    ordering = ("-to_date",)

    def get_queryset(self, request: HttpRequest) -> models.QuerySet["AccountStatementWithTransactionsCount"]:
        return (
            super()
            .get_queryset(request)
            .select_related("account")
            .annotate(transactions_count=models.Count("transactions"))
        )

    @admin.display(description=_("account"), ordering="account__name")
    def account_name(self, obj: AccountStatement) -> str:
        return obj.account.name

    transaction_changelist = reverse("admin:bankreader_transaction_changelist")

    @admin.display(description=_("transactions"), ordering="transactions_count")
    def transactions_link(self, obj: "AccountStatementWithTransactionsCount") -> str:
        return mark_safe(
            '<a href="{url}?account_statement__id__exact={account_statement_id}">{count}</a>'.format(
                url=self.transaction_changelist,
                account_statement_id=obj.id,
                count=obj.transactions_count,
            )
        )

    def save_model(
        self,
        request: HttpRequest,
        obj: AccountStatement,
        form: AccountStatementForm,
        change: bool,
    ) -> None:
        assert form.transactions is not None
        for message in obj.save_with_transactions(form.transactions):
            messages.warning(request, message)
        messages.success(request, _("Account statement was successfully loaded."))


@admin.register(Transaction)
class TransactionAdmin(ReadOnlyMixin, admin.ModelAdmin):
    date_hierarchy = "accounted_date"
    ordering = ("-accounted_date",)

    def get_list_display(self, request: HttpRequest) -> List[str | Callable[[Transaction], str]]:  # type: ignore
        return list(self.get_list_display_generator(request))

    def get_list_display_generator(
        self, request: HttpRequest
    ) -> Generator[str | Callable[[Transaction], str], None, None]:
        for field in Transaction._meta.fields[1:]:
            yield field.name

        for relation in get_transaction_relations().values():
            assert isinstance(relation.related_model, type(models.Model))
            RelatedModel: type[models.Model] = relation.related_model

            can_add = request.user.has_perm(f"{RelatedModel._meta.app_label}.add_{RelatedModel._meta.model_name}")
            can_see = request.user.has_perm(f"{RelatedModel._meta.app_label}.view_{RelatedModel._meta.model_name}")

            changelist_url = (
                reverse(
                    "admin:{}_{}_changelist".format(
                        RelatedModel._meta.app_label,
                        RelatedModel._meta.model_name,
                    )
                )
                if can_see
                else ""
            )

            add_url = (
                reverse(
                    "admin:{}_{}_add".format(
                        RelatedModel._meta.app_label,
                        RelatedModel._meta.model_name,
                    )
                )
                if can_add
                else ""
            )

            @admin.display(description=RelatedModel._meta.verbose_name)
            def related_object_link(obj: Transaction) -> str:
                try:
                    related_object = getattr(obj, relation.name)
                except RelatedModel.DoesNotExist:  # type: ignore
                    related_object = None
                if related_object:
                    return (
                        format_html(
                            '<a href="{changelist_url}?{remote_name}__id__exact={obj_id}">{text}</a>',
                            changelist_url=changelist_url,
                            remote_name=relation.remote_field.name,
                            obj_id=obj.id,
                            text=str(related_object),
                        )
                        if can_see
                        else str(related_object)
                    )
                elif can_add:
                    return format_html(
                        '<a href="{add_url}?{remote_name}={obj_id}" title="{title}">' '<img src="{icon}" alt="+"/></a>',
                        add_url=add_url,
                        remote_name=relation.remote_field.name,
                        obj_id=obj.id,
                        title=_("add"),
                        icon=static("admin/img/icon-addlink.svg"),
                    )
                return "-"

            yield related_object_link

    def get_list_filter(  # type: ignore
        self,
        request: HttpRequest,
    ) -> Tuple[str | Tuple[str, Type[admin.ListFilter]], ...]:
        base_filters: Tuple[str | Tuple[str, Type[admin.ListFilter]], ...] = (
            "account_statement__account",
            ("amount", AmountFieldListFilter),
        )
        related_filters: Tuple[str | Tuple[str, Type[admin.ListFilter]], ...] = tuple(
            (name, IdentifiedFieldListFilter) for name in get_transaction_relations()
        )
        return base_filters + related_filters

    def get_queryset(self, request: HttpRequest) -> models.QuerySet[Transaction]:
        return (
            super()
            .get_queryset(request)
            .select_related(
                "account",
                "account_statement",
                *(name for name in get_transaction_relations()),
            )
        )

    def has_add_permission(self, request: HttpRequest) -> bool:
        return False

    @admin.display(description=_("account statement"), ordering="account_statement__statement")
    def statement(self, obj: Transaction) -> str:
        return obj.account_statement.statement



import datetime
import decimal
from typing import IO, Iterable

from bankreader.models import Transaction

from .base import BaseReader


class BestReader(BaseReader):
    label = "Best"
    encoding = "cp1250"

    def read_transactions(self, statemen_file: IO) -> Iterable[Transaction]:
        rows = statemen_file.read().decode(self.encoding).split("\n")
        for row in rows:
            if row[:2] == "52":
                yield Transaction(
                    transaction_id=row[86:117].strip(),
                    entry_date=datetime.datetime.strptime(row[167:175], "%Y%m%d").date(),
                    accounted_date=datetime.datetime.strptime(row[175:183], "%Y%m%d").date(),
                    remote_account_number="%s-%s/%s"
                    % (
                        row[23:29],
                        row[29:39],
                        row[42:46] if row[39:42] == "000" else row[39:46],
                    ),
                    # 'remote_account_name': row[],
                    amount=decimal.Decimal("%s%s.%s" % ("-" if row[46] == "0" else "", row[50:63], row[63:65])),
                    variable_symbol=int(row[127:137]),
                    constant_symbol=int(row[137:147]),
                    specific_symbol=int(row[147:157]),
                    sender_description=row[269:409].strip(),
                    recipient_description=row[209:239].strip(),
                )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib import admin

from .models import Order, OrderPayment


@admin.register(Order)
class OrderAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "variable_symbol",
    )


@admin.register(OrderPayment)
class OrderPaymentAdmin(admin.ModelAdmin):
    list_display = ("order", "transaction", "amount")



from bankreader.readers import get_reader_choices
from django.apps.registry import Apps
from django.db import migrations, models
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
import django.db.models.deletion


def set_transaction_account_statement(apps: Apps, schema_editor: BaseDatabaseSchemaEditor) -> None:
    AccountStatement = apps.get_model("bankreader", "AccountStatement")
    Transaction = apps.get_model("bankreader", "Transaction")
    for transaction in Transaction.objects.iterator():
        transaction.account_statement = (
            AccountStatement.objects.filter(
                from_date__lte=transaction.accounted_date,
                to_date__gte=transaction.accounted_date,
            )
            .order_by("id")
            .first()
        )
        if transaction.account_statement:
            transaction.save()
    orphans = Transaction.objects.filter(account_statement=None)
    first, last = orphans.first(), orphans.last()
    if first and last:
        orphans.update(
            account_statement=AccountStatement.objects.create(
                account_id=transaction.account_id,
                statement="_deleted_",
                from_date=first.accounted_date,
                to_date=last.accounted_date,
            )
        )


class Migration(migrations.Migration):
    dependencies = [
        ("bankreader", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="accountstatement",
            options={
                "ordering": ("from_date",),
                "verbose_name": "account statement",
                "verbose_name_plural": "account statements",
            },
        ),
        migrations.AddField(
            model_name="account",
            name="reader",
            field=models.CharField(
                blank=True,
                choices=get_reader_choices(),
                max_length=150,
                null=True,
                verbose_name="account statement format",
            ),
        ),
        migrations.AddField(
            model_name="transaction",
            name="account_statement",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="transactions",
                to="bankreader.AccountStatement",
                verbose_name="account statement",
            ),
        ),
        migrations.RunPython(set_transaction_account_statement, reverse_code=migrations.RunPython.noop),
    ]



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("bankreader", "0002_transaction_account_statement"),
    ]

    operations = [
        migrations.AlterField(
            model_name="transaction",
            name="account_statement",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="transactions",
                to="bankreader.AccountStatement",
                verbose_name="account statement",
            ),
        ),
    ]



import re
from typing import IO, Iterable

from mt940.parser import parse as mt940_parse

from ..models import Transaction
from .base import BaseReader


class MT940Reader(BaseReader):
    label = "MT940 (MultiCash)"

    def read_transactions(self, statement_file: IO) -> Iterable[Transaction]:
        for transaction in mt940_parse(statement_file):
            purpose = transaction.data.get("purpose")
            symbols = dict(re.findall(r"([KVS]S) ([0-9]{10})", purpose))
            account_match = re.search(r"([0-9]+-)?([0-9]{10})/([0-9]{4})", purpose)
            description = re.split("([KVS]S [0-9]{10}){3}", purpose)[-1]
            yield Transaction(
                transaction_id=transaction.data.get("customer_reference"),
                entry_date=transaction.data.get("entry_date"),
                accounted_date=transaction.data.get("date"),
                remote_account_number=account_match.group() if account_match else "",
                remote_account_name=transaction.data.get("applicant_name"),
                amount=transaction.data.get("amount").amount,
                variable_symbol=symbols.get("VS"),
                constant_symbol=symbols.get("KS"),
                specific_symbol=symbols.get("SS"),
                sender_description=description,
                recipient_description=description,
            )



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DemoAppConfig(AppConfig):
    name = "bankreader_demo.demoapp"
    verbose_name = _("Bank Reader Demo App")

    def ready(self) -> None:
        # register bank statement readers
        from . import readers  # noqa



import datetime
from typing import IO, Iterable

from bankreader.models import Transaction

from .base import BaseReader


class GpcReader(BaseReader):
    label = "GPC"

    def read_transactions(self, statemen_file: IO) -> Iterable[Transaction]:
        rows = statemen_file.read().decode(self.encoding).split("\n")
        transaction = None
        for row in rows:
            # first row of transaction data
            if row[:3] == "075":
                if transaction:
                    # send previous transaction data
                    if "entry_date" not in transaction:
                        transaction["entry_date"] = transaction["accounted_date"]
                    yield Transaction(**transaction)
                    transaction = None
                if row[60] in "12":
                    # create new transaction data
                    amount = int(row[48:60]) / 100.0
                    transaction = {
                        "transaction_id": row[35:48],
                        "accounted_date": datetime.datetime.strptime(row[122:128], "%d%m%y").date(),
                        "remote_account_number": "%s-%s/%s"
                        % (
                            row[19:25],
                            row[25:35],
                            row[73:77],
                        ),
                        "remote_account_name": row[97:117].strip(),
                        "amount": amount if row[60] == "2" else -amount,
                        "variable_symbol": int(row[61:71]),
                        "constant_symbol": int(row[77:81]),
                        "specific_symbol": int(row[81:91]),
                    }
            # second row of transaction data
            elif transaction and row[:3] == "076":
                try:
                    transaction["entry_date"] = datetime.datetime.strptime(row[29:35], "%d%m%y").date()
                except ValueError:
                    transaction["entry_date"] = transaction["accounted_date"]
                transaction["sender_description"] = row[35:127].strip()
            # third row of transaction data
            elif transaction and row[:3] == "078":
                transaction["recipient_description"] = row[3:127].strip()
            # 4th row of transaction data
            elif transaction and row[:3] == "079":
                transaction["recipient_description"] += row[3:73].strip()
        if transaction:
            if "entry_date" not in transaction:
                transaction["entry_date"] = transaction["accounted_date"]
            yield Transaction(**transaction)



import os
import traceback
from typing import Any

from django.conf import settings
from django.core.management.base import BaseCommand, CommandParser
from django.db.models import Q

from ...models import Account, AccountStatement


class Command(BaseCommand):
    help = "Load Bank Statement"

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument("--account", dest="account", type=str)
        parser.add_argument("input_file", nargs="+", type=str)

    def handle(self, **options: Any) -> None:
        # get account
        try:
            q = Q(pk=int(options["account"]))
        except ValueError:
            q = Q(name=options["account"]) | Q(iban=options["account"])

        try:
            account = Account.objects.filter(reader__isnull=False).get(q)
        except Account.DoesNotExist:
            self.stderr.write(self.style.ERROR('Account "%s" does not exist' % options["account"]))
            return
        except Account.MultipleObjectsReturned:
            self.stderr.write(self.style.ERROR('Account "%s" is ambiguous' % options["account"]))
            return

        reader = account.get_reader()
        assert reader is not None

        for input_file in options["input_file"]:
            self.stdout.write(
                self.style.HTTP_INFO('Loading bank statement "%s" for account "%s"' % (input_file, account))
            )
            try:
                with open(input_file, "rb") as f:
                    transactions = tuple(reader.read_file(f))
            except Exception as e:
                if settings.DEBUG:
                    traceback.print_exc()
                self.stderr.write(self.style.ERROR('Error loading bank statement "%s": %s' % (input_file, e)))
                continue
            if not transactions:
                self.stderr.write(
                    self.style.ERROR('The account statement "%s" doesn\'t contain any transaction data.' % input_file)
                )
                continue
            statement = AccountStatement(account=account, statement=os.path.basename(input_file))
            messages = statement.save_with_transactions(transactions)
            for message in messages:
                self.stderr.write(self.style.WARNING(message))
            self.stdout.write(
                self.style.HTTP_INFO(
                    "Successfully loaded %d transactions (%d new) from %s."
                    % (len(transactions), len(transactions) - len(messages), input_file)
                )
            )



from django.db import migrations, models
import django.db.models.deletion
import localflavor.generic.models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Account",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(max_length=150, unique=True, verbose_name="account name"),
                ),
                (
                    "iban",
                    localflavor.generic.models.IBANField(
                        blank=True,
                        include_countries=None,
                        max_length=34,
                        null=True,
                        use_nordea_extensions=False,
                        verbose_name="IBAN",
                    ),
                ),
                (
                    "bic",
                    localflavor.generic.models.BICField(
                        blank=True, max_length=11, null=True, verbose_name="BIC (SWIFT)"
                    ),
                ),
            ],
            options={
                "verbose_name": "account",
                "verbose_name_plural": "accounts",
            },
        ),
        migrations.CreateModel(
            name="AccountStatement",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "statement",
                    models.CharField(max_length=256, verbose_name="statement"),
                ),
                (
                    "from_date",
                    models.DateField(editable=False, verbose_name="from date"),
                ),
                ("to_date", models.DateField(editable=False, verbose_name="to date")),
                (
                    "account",
                    models.ForeignKey(
                        limit_choices_to={"reader__isnull": False},
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="account_statements",
                        to="bankreader.Account",
                        verbose_name="account",
                    ),
                ),
            ],
            options={
                "ordering": ("from_date",),
            },
        ),
        migrations.CreateModel(
            name="Transaction",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "transaction_id",
                    models.CharField(max_length=256, verbose_name="transaction id"),
                ),
                ("entry_date", models.DateField(verbose_name="entry date")),
                ("accounted_date", models.DateField(verbose_name="accounted date")),
                (
                    "remote_account_number",
                    models.CharField(max_length=64, verbose_name="remote account number"),
                ),
                (
                    "remote_account_name",
                    models.CharField(max_length=128, verbose_name="remote account name"),
                ),
                (
                    "amount",
                    models.DecimalField(decimal_places=2, max_digits=20, verbose_name="amount"),
                ),
                (
                    "variable_symbol",
                    models.BigIntegerField(default=0, verbose_name="variable symbol"),
                ),
                (
                    "constant_symbol",
                    models.BigIntegerField(default=0, verbose_name="constant symbol"),
                ),
                (
                    "specific_symbol",
                    models.BigIntegerField(default=0, verbose_name="specific symbol"),
                ),
                (
                    "sender_description",
                    models.CharField(max_length=256, verbose_name="description for sender"),
                ),
                (
                    "recipient_description",
                    models.CharField(max_length=256, verbose_name="description for recipient"),
                ),
                (
                    "account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="bankreader.Account",
                        verbose_name="account",
                    ),
                ),
            ],
            options={
                "verbose_name": "transaction",
                "verbose_name_plural": "transactions",
                "ordering": ("accounted_date",),
            },
        ),
        migrations.AlterUniqueTogether(
            name="transaction",
            unique_together={("account", "transaction_id")},
        ),
    ]



from bankreader.readers import register_reader
from bankreader.readers.best import BestReader
from bankreader.readers.csv import CsvReader


@register_reader
class KbBestReader(BestReader):
    label = "Komerční banka Best"


@register_reader
class KbCsvReader(CsvReader):
    label = "Komerční banka CSV"
    encoding = "cp1250"
    delimiter = ";"
    column_mapping = {
        "Datum splatnosti": "accounted_date",
        "Datum odepsani JB": "entry_date",
        "Protiucet/Kod banky": "remote_account_number",
        "Nazev protiuctu": "remote_account_name",
        "Castka": "amount",
        "VS": "variable_symbol",
        "KS": "constant_symbol",
        "SS": "specific_symbol",
        "Identifikace transakce": "transaction_id",
        "Popis prikazce": "sender_description",
        "Popis pro prijemce": "recipient_description",
    }
    date_format = "%Y-%m-%d"
    decimal_separator = ","



from typing import IO, TYPE_CHECKING, Iterable
from zipfile import BadZipFile, ZipFile

if TYPE_CHECKING:
    from ..models import Transaction


class BaseReader:
    encoding = "utf-8"

    @property
    def label(self) -> str:
        raise NotImplementedError()

    def read_file(self, statement_file: IO) -> Iterable["Transaction"]:
        """Try to unpack ZIP archive and call read() for each file."""
        try:
            zip_file = ZipFile(statement_file)
        except BadZipFile:
            statement_file.seek(0)
            yield from self.read_transactions(statement_file)
        else:
            for zip_info in zip_file.filelist:
                with zip_file.open(zip_info) as f:
                    yield from self.read_file(f)

    def read_transactions(self, statement_file: IO) -> Iterable["Transaction"]:
        raise NotImplementedError()



import csv
import datetime
import decimal
import re
from typing import IO, Any, Dict, Iterable

from bankreader.models import Transaction

from .base import BaseReader


class CsvReader(BaseReader):
    label = "CSV"
    column_mapping: Dict[str, str] = {}
    date_format = "%Y-%m-%d"
    delimiter = ","
    quotechar = '"'
    encoding = "utf-8"
    decimal_separator = "."
    decimalregex = re.compile(r"[^0-9,-]")

    def __init__(self) -> None:
        self.decimal_cleaner = re.compile(r"[^0-9-%s]" % self.decimal_separator)

    def read_transactions(self, statemen_file: IO) -> Iterable[Transaction]:
        rows = statemen_file.read().decode(self.encoding).split("\n")
        column_mapping: Dict[str, int] = {}
        csv_reader = csv.reader(rows, delimiter=self.delimiter, quotechar=self.quotechar)
        for row in csv_reader:
            if not column_mapping:
                try:
                    column_mapping = {key: row.index(csv_key) for csv_key, key in self.column_mapping.items()}
                except ValueError:
                    pass
                continue
            yield Transaction(**{key: self.get_value(key, row[column_mapping[key]]) for key in column_mapping})

    def get_value(self, key: str, value: str) -> Any:
        if key in ("accounted_date", "entry_date"):
            return datetime.datetime.strptime(value, self.date_format).date()
        elif key == "amount":
            return decimal.Decimal(self.decimal_cleaner.sub("", value).replace(self.decimal_separator, "."))
        elif key.endswith("_symbol"):
            return int(value) if value.isdigit() else 0
        else:
            return value



# -*- coding: utf-8 -*-

from typing import Any

from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _

from bankreader.models import Transaction


class Order(models.Model):
    name = models.CharField(max_length=30)
    variable_symbol = models.CharField(max_length=10, unique=True)

    class Meta:
        verbose_name = _("order")
        verbose_name_plural = _("orders")

    def __str__(self) -> str:
        return self.name


class OrderPayment(models.Model):
    order = models.ForeignKey(Order, on_delete=models.CASCADE)
    transaction = models.OneToOneField(
        Transaction,
        editable=False,
        null=True,
        on_delete=models.CASCADE,
        related_name="identified_order_payment",
    )
    amount = models.DecimalField(_("amount"), decimal_places=2, max_digits=20)

    class Meta:
        verbose_name = _("order payment")
        verbose_name_plural = _("order payments")

    def __str__(self) -> str:
        return "{}, {}".format(self.order, self.amount)


@receiver(post_save, sender=Transaction)
def create_order_payment(instance: Transaction, **kwargs: Any) -> None:
    transaction = instance
    order = Order.objects.filter(variable_symbol=transaction.variable_symbol).first()
    if order:
        OrderPayment.objects.create(order=order, transaction=transaction, amount=transaction.amount)



from typing import TYPE_CHECKING, Dict, Type

if TYPE_CHECKING:
    from .base import BaseReader

readers: Dict[str, "BaseReader"] = {}


def get_reader_choices() -> list[tuple[str, str]]:
    return sorted([(key, reader.label) for key, reader in readers.items()], key=lambda x: x[1])


def register_reader(reader: Type["BaseReader"]) -> Type["BaseReader"]:
    readers[getattr(reader, "key", "%s.%s" % (reader.__module__, reader.__name__))] = reader()
    return reader
